// This is -*- C++ -*-
// $Id: checkfn.cc,v 1.6 1998/09/14 17:06:03 trow Exp $

/* checkfn.cc
 *
 * Copyright (C) 1998 EMC Capital Management, Inc.
 *
 * Developed by Jon Trowbridge <trow@emccta.com>.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 */


/*
  We check our special functions against numerical values that were
  calculated with Mathematica.
*/

#include <string>
#include <math.h>
#include "Random.h"
#include "statfn.h"

bool
mismatch(double fn, double ex)
{
  return isnan(fn) || (fabs(fn-ex) > 1e-6 && fabs(fn/ex-1) > 1e-6);
}

main()
{
  int p;

  cout << endl << "Checking factorial() and log_factorial()..." << endl;
  double f = 1;
  double lf = 0;
  for(int i=1; i<40; ++i) {
    f *= i;
    lf += log(i);
    double ff = factorial(i);
    double lff = log_factorial(i);
    if (mismatch(f, ff)) {
      cout << "factorial(" << i << ") evaluates to " << ff;
      cout << ", should equal " << f << endl;
    }
    if (mismatch(lf,lff)) {
      cout << "log_factorial(" << i << ") evaluates to " << lff;
      cout << ", should equal " << lf << endl;
    }
  }

  cout << endl << "Checking incomplete_beta()..." << endl;
  const double inc_beta_values[] = {
    2, 3, 1, 1,
    1, 2, 0.5, 0.75,
    2, 1, 0.75, 0.5625,
    5, 5, 0.4, 0.266568,
    0 };

  p=0;
  while (inc_beta_values[p]) {
    double a = inc_beta_values[p];
    double b = inc_beta_values[p+1];
    double x = inc_beta_values[p+2];
    double y = incomplete_beta(a,b,x);
    double val = inc_beta_values[p+3];
    if (mismatch(y,val)) {
      cout << "beta(" << a << "," << b << "," << x << ") ";
      cout << "evaluates to " << y << ", should equal " << val << endl;
    }
    p += 4;
  }

  cout << endl << "Checking log_gamma()..." << endl;
  const double log_gamma_values[] = {
       3,   0.6931471805599453094,
    M_PI,   0.8276945923234371015,
      10,  12.801827480081469611,
      17,  30.67186016080672804,
     101, 363.7393755555634901,
    0 };
  p=0;
  while (log_gamma_values[p]) {
    double x = log_gamma_values[p];
    double val = log_gamma_values[p+1];
    double lg = log_gamma(x);
    if (mismatch(lg, val)) {
      cout << "log_gamma(" << x << ") evaluates to " << lg;
      cout << ", should equal " << val << endl;
    }
    p += 2;
  }

  cout << endl << "Checking incomplete_gamma()..." << endl;
  const double inc_gamma_values[] = {
    3, 2, 0.323324,
    2, 3, 0.800852,
    6, 4, 0.21487,
    4, 6, 0.848796,
    9, 3, 0.00380299,
    3, 9,  0.993768,
    0 };

  p=0;
  while (inc_gamma_values[p]) {
    double a = inc_gamma_values[p];
    double x = inc_gamma_values[p+1];
    double val = inc_gamma_values[p+2];
    double ig = incomplete_gamma(a,x);
    if (mismatch(ig, val)) {
      cout << "incomplete_gamma(" << a << ", " << x << ") evaluates to " << ig;
      cout << ", should equal " << val << endl;
    }
    p += 3;
  }

  cout << endl << "Checking students_cdf()..." << endl;
  const double student_values[] = {
     7,  0.00, 0.5,
    99,  0.00, 0.5,
     3,  1.00, 0.8044988905221146790,
    20,  0.50, 0.6887340788594885,
    16, -0.50, 0.3119407776500231,
     1,  0.25, 0.5779791303773693,
     2, -3.00, 0.04773298313335456603,
     0
  };
  p=0;
  while (student_values[p]) {
    double df=student_values[p];
    double t=student_values[p+1];
    double val=student_values[p+2];
    double scdf = students_cdf(t,df);
    if (mismatch(scdf, val)) {
      cout << "students_cdf(" << t << ", " << df << ") evaluates to " << scdf;
      cout << ", should equal " << val << endl;
    }
    p += 3;
  }

  cout << endl << "Checking inv_students_cdf()..." << endl;
  for(int df=2; df<200; df += 3) {

    for (int ti=-30; ti<=30; ++ti) {
      double t = ti/10.0;
      double a1 = inv_students_cdf(students_cdf(t,df),df);
      if (mismatch(t,a1)) {
	cout << "inv_students_cdf(students_cdf(";
	cout << t << "," << df << ")," << df << ") = " << a1 << endl;
      }
    }

    for (double p=0.01; p <= 0.9905; p += 0.01) {
      double a1 = students_cdf(inv_students_cdf(p,df),df);
      if (mismatch(p,a1)) {
	cout << "students_cdf(inv_students_cdf(";
	cout << p << "," << df << ")," << df << ") = " << a1 << endl;
      }
    }
  }

  cout << endl << "Checking chi_square_cdf()..." << endl;
  const double chisq_values[] = {
      1,   2, 0.8427007929497148693,
      5, 3.2, 0.3308170979667568,
     10,  10, 0.5595067149347875886,
     20,   7, 0.003314944264632362892,
     60,  50, 0.18210391597745510980,
     80, 100, 0.9354296310788670242,
    120, 100, 0.09226505195893311030,
    122,   0, 0.0,
    0
  };
  p=0;
  while (chisq_values[p]) {
    double df=chisq_values[p];
    double x=chisq_values[p+1];
    double val=chisq_values[p+2];
    double csq=chi_square_cdf(x,df);
    if (mismatch(csq, val)) {
      cout << "chi_square_cdf(" << x << ", " << df << ") evaluates to " << csq;
      cout << ", should equal " << val << endl;
    }
    p += 3;
  }

  cout << endl << "Checking inv_chi_square_cdf()..." << endl;
  for (int df=2; df<200; df += 3) {

    for(double c2=df/4.0; c2 <= 2*df; c2 += 1.0) {
      // inv_chi_square_cdf() will have problems is p is too near 0 or 1...
      double pp = chi_square_cdf(c2,df);
      if (pp < 1e-5 || pp > 1-1e-5) continue;

      double a1 = inv_chi_square_cdf(chi_square_cdf(c2,df),df);
      if (mismatch(a1,c2)) {
	cout << "inv_chi_square_cdf(chi_square_cdf(";
	cout << c2 << "," << df << ")," << df << ") = " << a1 << endl;
      }
    }

    for (double p = 0.01; p <= 0.9905; p += 0.01) {
      double a1 = chi_square_cdf(inv_chi_square_cdf(p,df),df);
      if (mismatch(a1,p)) {
	cout << "chi_square_cdf(inv_chi_square_cdf(";
	cout << p << "," << df << "), " << df << ") = " << a1 << endl;
      }
    }
  }


}




// $Id: checkfn.cc,v 1.6 1998/09/14 17:06:03 trow Exp $


