// This is -*- C++ -*-
// $Id: DataImport.cc,v 1.6 1998/09/14 04:48:37 trow Exp $

/* DataImport.cc
 *
 * Copyright (C) 1998 EMC Capital Management, Inc.
 *
 * Developed by Jon Trowbridge <trow@emccta.com>.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 */

#include "DataImport.h"

#ifndef GOOSE_DEBUG
#define GOOSE_DEBUG
#endif

#ifdef GOOSE_DEBUG
static inline MIN(size_t a, size_t b) { return (a < b) ? a : b; }
#endif

// return true on success - exceptions would be nice.
static bool 
get_next_token(const string & separator,
	       const string & s,
	       // where in s to start looking 
	       size_t start,
	       // filled in.
	       // Index of next character after this token
	       size_t & token_end,
	       // Character after next separator; if no separator,
	       //  == token_end + 1
	       size_t & next_token)
{
#ifdef GOOSE_DEBUG
  printf("  ::Finding next token\n");
#endif

  if (start >= s.size())     return false;
  if (separator.size() == 0) return false;

#ifdef GOOSE_DEBUG
  printf("  ::start < s.size(), separator '%s', starting at '%s'...\n",
	  separator.c_str(), 
	  s.substr(start,MIN(start+10,s.size())-start).c_str());
#endif

  token_end = s.find_first_of(separator,start);

  if (token_end == string::npos) {

#ifdef GOOSE_DEBUG
    printf("  ::Didn't find a separator; returning entire string\n");
#endif

    // didn't find a separator, but this may be the last token.
    token_end  = s.size();
    next_token = token_end + 1; // on next call causes return false
  }
  else {
    next_token = token_end + separator.size();
  }

#ifdef GOOSE_DEBUG
  if (token_end > s.size()) {
    fprintf(stderr,"Token end %d too large!", token_end);
    return false;
  }
  printf("  ::Found token: '%s'\n", s.substr(start,token_end-start).c_str());
#endif

  return true;
}

static bool
parse_token(const DataType* dt, const string & s, double & result)
{
#ifdef GOOSE_DEBUG
  printf("Getting the value of token: %s\n", s.c_str());
#endif

  // remove leading and trailing whitespace, if any
  size_t i,j;
  for(i=0;i<s.size() && isspace(s[i]);++i);
  for(j=s.size()-1;j>=i && isspace(s[j]);--j);
  string ss = s.substr(i,j-i+1);

  if (ss.empty()) return false;

  if (dt != 0) {
    return dt->parse(ss,result);
  }
  // silly default
  else {
    result = atof(ss.c_str());
    return true;
  }
}

ImportEngine::~ImportEngine()
{
  for(int i=0; i<column_count(); ++i)
    delete column(i);

  if (results_not_requested_) {
    for(int i=0; i<results_.size(); ++i)
      delete results_[i];
  }
}


void 
ImportEngine::load(istream & is)
{
  // Fixme; we want this function, but also one that 
  //  tokenizes the stream on the fly. This would 
  //  be mutually exclusive with the use of the "guess" function.

  string text;

  char buf[4097];
  
  while (is) {
    is.read(buf,4096);
    buf[is.gcount()] = '\0'; 
    text += buf;
  }
  
  load(text);
}

void 
ImportEngine::load(const string & text)
{
#ifdef GOOSE_DEBUG
  if (results_.size() > 0) {
    fprintf(stderr,"ImportEngine::load() called twice?\n");
    return;
  }
#endif

  size_t i = 0;
  while (i < cols_.size()) {
    if (!cols_[i]->ignore()) {
      results_.push_back(new DataSet);
      results_[i]->set_label(cols_[i]->label());
    }
    // keep the indices on "results_" the same as cols_,
    //  just for convenience. trivial efficiency loss
    else results_.push_back(0);

    ++i;
  }

  size_t row  = 0;
  size_t pos  = 0;
  size_t end  = 0;
  size_t next = 0;
  while ( get_next_token(sep_,
			 text,
			 pos,
			 end,
			 next) ) {

#ifdef GOOSE_DEBUG
    printf("--->row: %d pos: %d end: %d next: %d size: %d\n",
	    row,pos,end,next,text.size());
    printf("--->Entering column-parsing loop\n");
#endif

    size_t col = 0;
    size_t cpos = 0;
    size_t cend = 0;
    size_t cnext = 0;
    string rowtext = text.substr(pos,end-pos);
    while ( get_next_token(column(col)->separator(),
			   rowtext,
			   cpos,
			   cend,
			   cnext) ) {

      // skip it if we've reached the max or it's an
      //  ignore.
      if ( cols_[col]->ignore() ||
	   cols_[col]->has_max() && (row > cols_[col]->max_elements())) {
	cpos = cnext;
	++col;
	continue;
      }

#ifdef GOOSE_DEBUG
      printf("--->col: %d cpos: %d cend: %d cnext: %d size: %d\n",
	     col,cpos,cend,cnext,rowtext.size());
#endif
      
      // reserve space in blocks
      // should happen on 0, 256, etc.
      if (row % 256 == 0) {
	results_[col]->reserve(row+1);
      }

      double value;
      if ( parse_token(cols_[col]->type(),
		       rowtext.substr(cpos,cend-cpos),
		       value) ) {
	results_[col]->add(value);
      }
      
      cpos = cnext;
      ++col;
    }

#ifdef GOOSE_DEBUG
    printf("--->Leaving column-parsing loop\n");
#endif

    pos = next;
    ++row;
  }
#ifdef GOOSE_DEBUG
  printf("--->Parsed %d rows\n", 
	 row);
#endif

  results_not_requested_ = true;
}


#if 0
/////////////////////////////////////
///// Why this code isn't used now:
///// 1) it requires us to know a priori what the row separators are,
/////    in order to put them in the vector
///// 2) we can't go over a stream multiple times, but in general 
/////    we need to make multiple passes over the data for guessing
/////    its nature and GUI feedback. However, going only once is 
/////    best for non-interactive use; so the new code will have to do
/////    both
///// 3) I would like to eventually build a data structure describing 
/////    the parsing; e.g. a table of token locations in the text 
/////    being parsed. Thus it's better to do the parsing on one
/////    large string with indices. The data structure will be used 
/////    for guessing at the columns and user feedback (e.g. a text widget
/////    with tokens highlighted, something like that)

// pull out token for this column,
//  leaving other tokens in s.
string
ColumnSpec::extract(string& s)
{
#if 0
  // I don't understand this part
  char term[3] = "  ";
  if (separator())
    term[1] = separator();
#endif

  cerr << "Extracting" << endl;

  size_t len = s.size();
  size_t beg = 0;
  while (true) {
    if (beg == len) {
      // s was just whitespace or ""
      s = "";
      return "";
    }
    // Need to use isspace for i18n and to get tabs, etc.
    // It's OK if the separator is whitespace, because this
    //  is before the token; the separator is after the token.
    if (isspace(s[beg])) ++beg;
    else break;
  }
  // at loop end beg should index the first non-whitespace character,
  //  and beg != s.npos since we would have returned.
  cerr << "Whitespace cleared." << endl;


  size_t end = s.find_first_of(sep_, beg);

  // This can only happen (I hope) if the field was empty (i.e. two terminator
  // characters were separated only by whitespace)
  // no need for the if - things just work, ex becomes ""
  //  if (end == beg) end = beg;

  cerr << "Begin: " << beg << " End: " << end << endl;
   
  string ex;
  if (end == s.npos) {
    // this will happen on the last column, with no separator after it.
    ex = s.substr(beg);
    s = "";
  } else {
    ex = s.substr(beg, end-beg);

    // skip the separator, but leave any whitespace for next call.
    end += sep_.size();
    // end can't be npos unless the string has length 
    //  UINT_MAX because we just checked that above.
    s = s.substr(end);
  }

  cerr << "Returning token: `" << ex << "'" << endl;

  return ex;
}

//////////////////////////////////////////////////////////////////////////////

ImportEngine::~ImportEngine()
{
  if (dealloc_strvec_)
    delete strvec_;

  for(int i=0; i<column_count(); ++i)
    delete column(i);
}

void
ImportEngine::preload_istream(istream& in)
{
  cerr << "Preloading istream" << endl;

  dealloc_strvec_ = true;
  strvec_ = new vector<string>;

  while ( !in.eof() && in.good() ) {
    string next = get_line(&in);
    cerr << "Preloading line: `" << next << "'" << endl;
    
    // ignore blank lines.
    if (next.length() > 0) strvec_->push_back(next);
  }

  strveciter_ = strvec_->begin();
}

vector<DataSet*>
ImportEngine::import()
{
  vector<DataSet*> dsvec;
  // construct our vector of empty DataSet objs
  for(int i=0; i<column_count(); ++i) {
    if (! column(i)->ignore()) {
      DataSet* ds = new DataSet;
      ds->set_label(column(i)->label());
      if (row_count() != -1)
	ds->reserve(row_count());
      dsvec.push_back(ds);
    }
  }

  while (ready()) {
    cerr << "About to get next line." << endl;

    string s = next_line();

    cerr << "Next line: `" << s << "'" << endl;

    if (s.length()) { // skip empty lines

      // We don't do anything clever in the event of "short lines".  Our error
      // handling generally sucks.
      vector<DataSet*>::iterator dsi = dsvec.begin();
      int i = 0;
      while ( (i<column_count()) && (dsi != dsvec.end()) ) {

	string ex = column(i)->extract(s);

	if (! column(i)->ignore()) {
	  const DataType* t = column(i)->type();
	  double x; 

	  if (t != 0) x = t->parse(ex);
	  else x = atof(ex.c_str()); // just pretend it's a float

	  cerr << "Adding value: `" << x << "'" << endl;

	  (*dsi)->add(x);
	  ++dsi;
	  ++i;
	}
      }
      cerr << "Finished parsing line." << endl;

    }
  }

  return dsvec;
}

bool
ImportEngine::ready()
{
  bool r;

  if (in_) {
    r = in_->good() && !in_->eof();
    cerr << "Istream goodness: " << r << endl;
  } else if (strvec_) {
    r = strveciter_ != strvec_->end();
    cerr << "strvec at end?: " << r << endl;
  } else {
    r = false;
    cerr << "Assuming not ready" << endl;
  }

  return r;
}


// Returns "" when there's not another line or the line is
//  empty.
string 
ImportEngine::get_line(istream* is)
{
  cerr << "get_line" << endl;

  if (is == 0) return "";

  string line("");

  while ( true ) {
    static const size_t chunk = 2048;
    char buffer[chunk];

    while ( true ) {
      if (is->eof() || !is->good()) {
	cerr << "EOF reached; top of loop" << endl;
	break;
      }
	
      is->get(buffer, chunk, sep_[0]);
      
      line += buffer;

      cerr << "Adding buffer contents: `" << buffer << "'" << endl;

      if ( is->gcount() == chunk ) {
	// make sure we didn't stop on a separator start
	char c;
	is->get(c);
	cerr << "Peeking at `" << c << "'" << endl;
	if (c == sep_[0]) {
	  is->putback(c);
	  cerr << "Put back `" << c << "'" << endl;
	  break;
	}
	// no separator start, keep reading,
	//  no need to gobble the terminator.
      }
      else break; // done.
    }
      
    if (is->eof() || !is->good()) {
      // line is just whatever was left of the file,
      // a separator is not required at the end.
      cerr << "EOF reached; after loop" << endl;
      break;
    }

    // OK, we may not have a real separator,
    //  but the first char in cin should 
    //  be the first char in sep_
    const size_t sepsize = sep_.size();

    // is this a gcc extension? stroustrup 
    //  says a constant expression is OK
    char sepbuf[sepsize+1];

    // have to use read because we don't know any specific 
    //  terminator.
    is->read(sepbuf, sepsize);
    sepbuf[is->gcount()] = '\0';

    cerr << "Potential separator: `" << sepbuf << "'" << endl;

    if (is->gcount() < sepsize) {
      // couldn't be a separator, since it's too short; also, we're at
      // the the end of the file. So just add whatever non-separator
      // thing it was, gobble EOF, and finish.
      line += sepbuf;

      char c = is->get();
      // debug
      if (c != EOF) cerr << "Gobbled `" << c << "' rather than EOF" << endl;
      break;
    }
    else if (sepbuf == sep_) {
      // It is a separator; discard it implicitly, 
      //  and we're done.
      cerr << "Discarded separator: `" << sepbuf << "'" << endl;
      //      char c = is->get();
      break;
    }
    else {
      // wasn't a separator, so append everything and keep going.
      line += sepbuf;
    }
  }

  cerr << "Got line: `" << line << "'" << endl;

  //  If this is the end of the file we still want to 
  //   return the line we just sucked in, no? 
  //   The next ready() call will grab the end of file
  //    return in_->eof() ? string("") : string(buffer);
  return line;  
}

string
ImportEngine::next_line()
{
  ++row_;

  if (in_) {
    cerr << "Getting next line from istream" << endl;
    return get_line(in_);
  } else if (strvec_) {
    cerr << "Getting next line from string vector" << endl;
    return *(strveciter_++);
  } else {
    // this shouldn't happen, right?
    cerr << "No istream or string vector" << endl;
    return string("");
  }
}

#endif

// $Id: DataImport.cc,v 1.6 1998/09/14 04:48:37 trow Exp $

