// This is -*- C++ -*-
// $Id: DataSet.h,v 1.14 1998/09/14 04:48:37 trow Exp $

/* 
 * DataSet.h
 *
 * Copyright (C) 1998 EMC Capital Management, Inc.
 *
 * Developed by Jon Trowbridge <trow@emccta.com>.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 */

#ifndef _INC_DATASET_H
#define _INC_DATASET_H

#include <iostream.h>
#include <nan.h>
#include <math.h>
#include <string>
#include "RCArray.h"
#include "Random.h"
#include "DataType.h"

class DataSet {
public:
  DataSet();

  DataSet(const DataSet&);
  ~DataSet() { }

  DataSet& operator=(const DataSet&);

  // type isn't acutally used right now.
  const DataType* type() const { return type_; }
  void set_type(const DataType& t) { type_ = &t; }

  const string& label() const { return label_; }
  void set_label(const string& s) { label_ = s; }

  size_t size() const { return data_.size(); }
  const double* data() const { return data_.data(); }
  // note: no bounds checking
  double data(size_t i) const { return (data_.data())[i]; } 

  const double* sorted_data() const;
  double sorted_data(size_t i) const { return sorted_data()[i]; }
  
  double min() const { return min_; }
  double max() const { return max_; }
  double range() const { return max_ - min_; }

  double sum() const { return sum_; }
  double mean() const { return size() ? sum_/size() : NAN; }
  double var() const { return size() ? sumsq_/size() : NAN; }
  double sdev() const { return size() ? sqrt(sumsq_/size()) : NAN; }
  double sdevs() const { return size()>1 ? sqrt(sumsq_/(size()-1)) : NAN; }

  double percentile(double p) const;
  double med() const { return percentile(0.5); }
  double q1() const { return percentile(0.25); }
  double q3() const { return percentile(0.75); }
  double iqr() const { return q3() - q1(); } // interquartile range
  double decile(int i) const { return (i<1||i>9) ? NAN : percentile(i/10.0); }

  // There are some computed (non-cached) descriptive statistics.  They are
  // a bit more obscure, so I don't bother keeping a running tally of them
  // via add().  (Or, in some cases, I can't...)
  double mean_trimmed(size_t left, size_t right) const;
  double mean_trimmed(size_t t) const { return mean_trimmed(t,t); }
  double mean_winsorized(size_t left, size_t right) const;
  double mean_winsorized(size_t t) const { return mean_winsorized(t,t); }
  double moment(unsigned k, double x) const; // k-th moment about x
  double moment(unsigned k) const { return moment(k,mean()); }
  double gmean() const;    // geometric mean
  double hmean() const;    // harmonic mean
  double rms() const;      // root mean square
  double meandev() const;  // mean deviation
  double meddev() const;   // median deviation
  double kurtosis() const; // kurtosis
  double skewness() const; // skewness

  double excess_kurtosis() const { return kurtosis()-3; }
  double momental_skewness() const { return skewness()/2; }

  // Here are some functions of two DataSets
  double covar(const DataSet& ds) const;
  double corr(const DataSet& ds) { return covar(ds)/(sdev()*ds.sdev()); }
  double pooled_mean(const DataSet& ds) const;
  double pooled_var(const DataSet& ds) const;
  double weighted_mean(const DataSet& wt) const;

  int sortedness() const { return sortedness_; }
  bool increasing() const { return sortedness_ == 1; }
  bool decreasing() const { return sortedness_ == -1; }
  bool unordered() const { return sortedness_ == 0; }
  bool constant() const { return sortedness_ == 2; }

  size_t greater_than(double x) const;
  size_t less_than(double x) const ;
  size_t between(double a, double b) const {
    return size()-less_than(a)-greater_than(b);
  }
  size_t equal_to(double x) const {
    return size()-less_than(x)-greater_than(x);
  }

  // make sure we have allocated room for at least R entries
  void reserve(size_t R); 

  void add(double);

  void clear();

  void append(const DataSet&);

  void apply(double (*)(double));
  template<class FuncObj>
  void apply(FuncObj f) {
    for(size_t i=0; i<size(); ++i)
      data_[i] = f(data_[i]);
    recalc();
  }

  // transform by x -> ax+b
  void linear_transform(double a, double b);
  // transform by x -> log(x) (natural log, of course...)
  void log_transform();
  void exp_transform();
  void logit_transform();

  void reverse();
  void sort(int dir = 1);
  void scramble();
  void rank();

  // useful for debugging, too gross for much else
  void spew(ostream& out, char sep = ' ') const {
    const double* d = data();
    for(size_t i=0; i<size(); ++i) out << d[i] << sep;
    if (sep != '\n')
      out << endl;
  }

  size_t random_index() const {
    // More stupid const tricks
    return (size_t)((DataSet*)this)->rnd.random_integer(0,size()-1);
  }
  double random_element() const { return data(random_index()); }
  void random_seed(unsigned s) { rnd.seed(s); }

  double epsilon() const { return epsilon_; }
  void set_epsilon(double e) { epsilon_ = e; }

private:
  void recalc();
  void recalc_saving_sorted();
  void copy(const DataSet&);

  // convenience routines...
  bool is_zero(double x) const { return fabs(x) < epsilon_; }
  bool are_equal(double x, double y) const { return is_zero(x-y); }
  bool is_positive(double x) const { return x > epsilon_; }

  string label_;

  RCArray<double> data_;
  RCArray<double> sdata_;

  double min_, max_, sum_;
  // If we call our data values x_1,...,x_N, then sumsq_ contains
  // \sum_{i=0}^N(x_i-\mu)^2, where \mu is the mean of the x_i.
  double sumsq_;

  // sortedness == +1 if data is increasing
  //            == -1 if data is decreasing
  //            ==  0 if data is unordered
  //            == +2 if the data is constant (a bad hack)
  int sortedness_;

  double epsilon_;

  Random rnd;

  const DataType* type_;
};

#endif // _INC_DATASET_H

// $Id: DataSet.h,v 1.14 1998/09/14 04:48:37 trow Exp $
