// This is -*- C++ -*-
// $Id: matrix.cc,v 1.2 1998/09/14 04:48:38 trow Exp $

/* matrix.cc
 *
 * Copyright (C) 1998 EMC Capital Management, Inc.
 *
 * Developed by Jon Trowbridge <trow@emccta.com>.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 */


#include <math.h>
#include <iomanip.h>
#include "matrix.h"

inline
double& mat(double* m, int N, int i, int j)
{
  return m[i*N+j];
}

void
matrix_print(ostream& out, double* m, int N)
{
  for(int i=0; i<N; ++i) {
    for(int j=0; j<N; ++j)
      cout << setw(9) << mat(m,N,i,j) << " ";
    cout << endl;
  }
}

bool
matrix_invert(double* m, int N)
{
  double* scale = new double[N];
  int* perm = new int[N];
  
  // First, compute scaling values
  for(int i=0; i<N; ++i) {
    double max = 0.0;
    for(int j=0; j<N; ++j) {
      double x = fabs(mat(m,N,i,j));
      if (x > max) max = x;
    }
    if (max == 0.0) return false;
    scale[i] = 1/max;
  }

  int i_max=-1;
      
  // Next, use Crout's method to do an LU decomposition
  for(int j=0; j<N; ++j) {

    for(int i=0; i<j; ++i) {
      double sum = mat(m,N,i,j);
      for(int k=0; k<i; ++k)
	sum -= mat(m,N,i,k) * mat(m,N,k,j);
      mat(m,N,i,j) = sum;
    }

    double max = 0.0;
    for(int i=j; i<N; ++i) {
      double sum = mat(m,N,i,j);
      for(int k=0; k<j; ++k)
	sum -= mat(m,N,i,k) * mat(m,N,k,j);
      mat(m,N,i,j) = sum;
      double x = scale[i] * fabs(sum);
      if (x >= max) {
	max = x;
	i_max = i;
      }
    }

    if (j != i_max) {
      // Interchange rows
      for(int k=0; k<N; ++k) {
	double t = mat(m,N,i_max,k);
	mat(m,N,i_max,k) = mat(m,N,j,k);
	mat(m,N,j,k) = t;
      }
      scale[i_max] = scale[j];
    }

    perm[j] = i_max;

    if (j != N-1) {
      double piv = mat(m,N,j,j);
      if (piv == 0.0) return false;
      piv = 1/piv;
      for(int i=j+1; i<N; ++i)
	mat(m,N,i,j) *= piv;
    }
  }

  delete [] scale;

  // Now we solve for the inverse, column by column
  double* inv = new double[N*N];
  double* col = new double[N];
  
  for(int j=0; j<N; ++j) {
    for(int i=0; i<N; ++i)
      col[i] = i == j ? 1.0 : 0.0;
    
    int ii = -1;
    for(int i=0; i<N; ++i) {
      int ip = perm[i];
      double sum = col[ip];
      col[ip] = col[i];
      if (ii >= 0)
	for(int k=ii; k<i; ++k)
	  sum -= mat(m,N,i,k) * col[k];
      else if (sum)
	ii = i;
      col[i] = sum;
    }

    for(int i=N-1; i>=0; --i) {
      double sum = col[i];
      for(int k=i+1; k<N; ++k)
	sum -= mat(m,N,i,k) * col[k];
      col[i] = sum/mat(m,N,i,i);
    }

    for(int i=0; i<N; ++i)
      mat(inv,N,i,j) = col[i];
  }

  delete [] col;

  // Copy our inverse into our original matrix array
  memcpy(m, inv, sizeof(double)*N*N);
  delete [] inv;

  delete [] perm;

  return true;
}




// $Id: matrix.cc,v 1.2 1998/09/14 04:48:38 trow Exp $
