// This is -*- C++ -*-
// $Id: statfn.cc,v 1.7 1998/09/14 04:48:38 trow Exp $


/* statfn.cc
 *
 * Copyright (C) 1998 EMC Capital Management, Inc.
 *
 * Developed by Jon Trowbridge <trow@emccta.com>.
 *
 * Parts of this code are (C) 1998 Ross Ihaka, and is derived from his
 * Mathlib code that is distributed with R (http://fangorn.ci.tuwien.ac.at/R).
 * Functions that originated in R are labelled below.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 */

#include <iostream.h>
#include <math.h>
#include <nan.h>
#include "statfn.h"

double
factorial(int n)
{
  static double lookup[24] = {
    1.,
    1.,
    2.,
    6.,
    24.,
    120.,
    720.,
    5040., 
    40320.,
    362880.,
    3628800.,
    39916800.,
    479001600.,
    6227020800.,
    87178291200.,
    1307674368000.,
    20922789888000.,
    355687428096000.,
    6402373705728000.,
    121645100408832000.,
    2432902008176640000.,
    51090942171709440000.,
    1124000727777607680000.,
    25852016738884976640000.
  };
  if (n < 0)
    return NAN;
  else if (n < 24)
    return lookup[n];
  else
    return exp(log_factorial(n));
}

double
log_factorial(int n)
{
  return n >= 0 ? log_gamma(1.0+n) : NAN;
}

double
choose(int n, int k)
{
  if (n < 1 || k < 0 || k > n)
    return 0;
  else if (n < 13)
    return factorial(n)/(factorial(k)*factorial(n-k));
  else
    return exp(log_choose(n,k));
}

double
log_choose(int n, int k)
{
  return log_factorial(n) - log_factorial(k) - log_factorial(n-k);
}

double
log_gamma(double x)
{
  double sum = 1.000000000190015
    + 76.18009172947146 / (x+1)
    - 86.50532032941677 / (x+2)
    + 24.01409824083091 / (x+3)
    - 1.231739572450155 / (x+4)
    + 0.1208650973866179e-2 / (x+5)
    - 0.5395239384953e-5 / (x+6);

  return log(2.5066282746310005 * sum / x) - x - 5.5 + (x + 0.5)*log(x + 5.5);
}

// This function is adapted from one in R (see license header above)
double
incomplete_gamma(double p, double x)
{
  const double plimit = 1000.0;
  const double elimit = -88.0;
  const double xbig = 1.0e+8;
  const double EPSILON = 1e-16;
  const double oflo = 1.0e+37;

  double pn1, pn2, pn3, pn4, pn5, pn6, arg, c, rn, a, b, an;
  double sum;

  /* check that we have valid values for x and p */
  
  if (isnan(x) || isnan(p) || p <= 0.0)
    return NAN;

  if (x <= 0.0)
    return 0.0;

  /* use a normal approximation if p > plimit */

  if (p > plimit) {
    pn1 = sqrt(p) * 3.0 * (pow(x/p, 1/3.0) + 1.0 / (p * 9.0) - 1.0);
    return normal_cdf(pn1);
  }

  /* if x is extremely large compared to p then return 1 */

  if (x > xbig)
    return 1.0;

  if (x <= 1.0 || x < p) {

    /* use pearson's series expansion. */
    
    arg = p * log(x) - x - log_gamma(p + 1.0);
    c = 1.0;
    sum = 1.0;
    a = p;
    do {
      a = a + 1.0;
      c = c * x / a;
      sum = sum + c;
    } while (c > EPSILON);
    arg = arg + log(sum);
    sum = 0.0;
    if (arg >= elimit)
      sum = exp(arg);
  } else {
    
    /* use a continued fraction expansion */

    arg = p * log(x) - x - log_gamma(p);
    a = 1.0 - p;
    b = a + x + 1.0;
    c = 0.0;
    pn1 = 1.0;
    pn2 = x;
    pn3 = x + 1.0;
    pn4 = x * b;
    sum = pn3 / pn4;
    for (;;) {
      a = a + 1.0;
      b = b + 2.0;
      c = c + 1.0;
      an = a * c;
      pn5 = b * pn3 - an * pn1;
      pn6 = b * pn4 - an * pn2;
      if (fabs(pn6) > 0.0) {
	rn = pn5 / pn6;
	if (fabs(sum - rn) <= (rn<1 ? rn*EPSILON : EPSILON))
	  break;
		
	sum = rn;
      }
      pn1 = pn3;
      pn2 = pn4;
      pn3 = pn5;
      pn4 = pn6;
      if (fabs(pn5) >= oflo) {
	
	/* re-scale the terms in continued fraction */
	/* if they are large */

	pn1 = pn1 / oflo;
	pn2 = pn2 / oflo;
	pn3 = pn3 / oflo;
	pn4 = pn4 / oflo;
      }
    }
    arg = arg + log(sum);
    sum = 1.0;
    if (arg >= elimit)
      sum = 1.0 - exp(arg);
  }
  return sum;
}

// This function is adapted from one in R (see license header above)
double
incomplete_beta(double pin, double qin, double x)
{
  if (isnan(x) || isnan(pin) || isnan(qin) || pin <= 0.0 || qin <= 0.0)
    return NAN;
  
  if (x <= 0)
    return 0;
  if (x >= 1)
    return 1;

  double ans, c, finsum, p, ps, p1, q, term, xb, xi, y;
  int n, i, ib;

  const double eps = 1e-16;
  static double alneps = log(eps);
  static double sml = DBL_MIN;
  static double alnsml = log(sml);

  y = x;
  p = pin;
  q = qin;

  /* swap tails if x is greater than the mean */

  if (p / (p + q) < x) {
    y = 1 - y;
    p = qin;
    q = pin;
  }

  if ((p + q) * y / (p + 1) < eps) {

    /* tail approximation */

    ans = 0;
    xb = p * log(y<sml ? sml : y) - log(p) - log_beta_function(p, q);
    if (xb > alnsml && y != 0)
      ans = exp(xb);
    if (y != x || p != pin)
      ans = 1 - ans;
  }
  else {

    /* evaluate the infinite sum first.  term will equal */
    /* y^p / beta(ps, p) * (1 - ps)-sub-i * y^i / fac(i) */

    ps = q - floor(q);
    if (ps == 0)
      ps = 1;
    xb = p * log(y) - log_beta_function(ps, p) - log(p);
    ans = 0;
    if (xb >= alnsml) {
      ans = exp(xb);
      term = ans * p;
      if (ps != 1) {
	n = (int)(alneps/log(y));
	if (n < 4) n = 4;
	for(i=1 ; i<= n ; i++) {
	  xi = i;
	  term = term * (xi - ps) * y / xi;
	  ans = ans + term / (p + xi);
	}
      }
    }

    /* now evaluate the finite sum, maybe. */
    
    if (q > 1) {
      xb = p * log(y) + q * log(1 - y) - log_beta_function(p, q) - log(q);
      ib = (int)(xb/alnsml);
      if (ib < 0) ib = 0;
      term = exp(xb - ib * alnsml);
      c = 1 / (1 - y);
      p1 = q * c / (p + q - 1);

      finsum = 0;
      n = (int)q;
      if (q == n)
	n = n - 1;
      for(i=1 ; i<=n ; i++) {
	if (p1 <= 1 && term / eps <= finsum)
	  break;
	xi = i;
	term = (q - xi + 1) * c * term / (p + q - xi);
	if (term > 1) {
	  ib = ib - 1;
	  term = term * sml;
	}
	if (ib == 0)
	  finsum = finsum + term;
      }
      ans = ans + finsum;
    }
    if (y != x || p != pin)
      ans = 1 - ans;
    if (ans > 1.0) ans = 1.0;
    if (ans < 0.0) ans = 0.0;
  }
  return ans;
}

// This function is adapted from one in R (see license header above)
double
inv_gamma_cdf(double p, double alpha, double scale)
{
  const int MAXIT = 20;

  const double C1 = 0.01;
  const double C2 = 0.222222;
  const double C3 = 0.32;
  const double C4 = 0.4;
  const double C5 = 1.24;
  const double C6 = 2.2;
  const double C7 = 4.67;
  const double C8 = 6.66;
  const double C9 = 6.73;

  const double C10 = 13.32;
  const double C11 = 60.0;
  const double C12 = 70.0;
  const double C13 = 84.0;
  const double C14 = 105.0;
  const double C15 = 120.0;
  const double C16 = 127.0;
  const double C17 = 140.0;
  const double C18 = 1175.0;
  const double C19 = 210.0;

  const double C20 = 252.0;
  const double C21 = 2264.0;
  const double C22 = 294.0;
  const double C23 = 346.0;
  const double C24 =420.0;
  const double C25 = 462.0;
  const double C26 = 606.0;
  const double C27 = 672.0;
  const double C28 = 707.0;
  const double C29 = 735.0;

  const double C30 = 889.0;
  const double C31 = 932.0;
  const double C32 = 966.0;
  const double C33 = 1141.0;
  const double C34 = 1182.0;
  const double C35 = 1278.0;
  const double C36 = 1740.0;
  const double C37 = 2520.0;
  const double C38 = 5040.0;

  double a, b, c, ch, g, p1, v;
  double p2, q, s1, s2, s3, s4, s5, s6, t, x, xx;
  double aa = 0.6931471806;
  double e = 0.5e-6;
  double pmin = 0.000002;
  double pmax = 0.999998;
  int i;
  
  /* test arguments and initialise */
  
  if (isnan(p) || isnan(alpha) || isnan(scale) ||
      p < pmin || p > pmax || alpha <= 0)
    return NAN;
  
  v = 2*alpha;
  
  /* xx = 0.5*v; */
  
  xx = alpha;
  c = xx-1.0;
  g = lgamma(0.5*v);
    
  if(v < (-C5)*log(p)) {

    /* starting approximation for small chi-squared */
    
    ch = pow(p*xx*exp(g+xx*aa),1.0/xx);
    if(ch < e) 
      return NAN;
  } else if (v > C3) {

    /* starting approximation using Wilson and Hilferty estimate */
    
    x = inv_normal_cdf(p);
    p1 = C2/v;
    ch = v*pow(x*sqrt(p1)+1.0-p1, 3.0);
    
    /* starting approximation for p tending to 1 */
    
    if( ch>C6*v+6.0 )
      ch = -2.0*(log(1.0-p)-c*log(0.5*ch)+g);
  }
  else {
    
    /* starting approximation for v less than or equal to 0.32 */
    
    ch = C4;
    a = log(1.0-p);
    do {
      q = ch;
      p1 = 1.0+ch*(C7+ch);
      p2 = ch*(C9+ch*(C8+ch));
      t = -0.5+(C7+2.0*ch)/p1-(C9+ch*(C10+3.0*ch))/p2;
      ch = ch-(1.0-exp(a+g+0.5*ch+c*aa)*p2/p1)/t;
    } while(fabs(q/ch-1.0) > C1);
  }
  
  /* algorithm as 239 and calculation of seven term taylor series */
  
  for( i=1 ; i<=MAXIT  ; i++ ) {
    q = ch;
    p1 = 0.5*ch;
    p2 = p - incomplete_gamma(xx, p1);
    if(!finite(p2))
      return NAN;
    
    t = p2*exp(xx*aa+g+p1-c*log(ch));
    b = t/ch;
    a = 0.5*t-b*c;
    s1 = (C19+a*(C17+a*(C14+a*(C13+a*(C12+C11*a)))))/C24;
    s2 = (C24+a*(C29+a*(C32+a*(C33+C35*a))))/C37;
    s3 = (C19+a*(C25+a*(C28+C31*a)))/C37;
    s4 = (C20+a*(C27+C34*a)+c*(C22+a*(C30+C36*a)))/C38;
    s5 = (C13+C21*a+c*(C18+C26*a))/C37;
    s6 = (C15+c*(C23+C16*c))/C38;
    ch = ch+t*(1.0+0.5*t*s1-b*c*(s1-b*(s2-b*(s3-b*(s4-b*(s5-b*s6))))));
    if(fabs(q/ch-1.0) > e)
      return 0.5*scale*ch;
  }
  return 0.5*scale*ch;
}



//////////////////////////////////////////////////////////////////////////

double
log_beta_function(double z, double w)
{
  return log_gamma(z)+log_gamma(w)-log_gamma(z+w);
}

double
beta_function(double z, double w)
{
  return exp(log_gamma(z)+log_gamma(w)-log_gamma(z+w));
}


/////////////////////////////////////////////////////////////////////////////

double
normal_cdf(double x)
{
  x /= M_SQRT2;
  return 0.5 * (x >= 0 ? 1 + erf(x) : erfc(-x));
}

double
inv_normal_cdf(double p)
{
  double x = p;

  int sign;
  if (x < 0.5) {
    sign = -1;
  } else {
    sign = 1;
    x = 1-x;
  }

  if (x < 1e-20) return sign*10;

  x = sqrt(-2*log(x));    
  
  const double p0 = -0.322232431088;
  const double p1 = -1.0;
  const double p2 = -0.342242088547;
  const double p3 = -0.0204231210245;
  const double p4 = -0.453642210148e-4;

  const double q0 = 0.0993484626060;
  const double q1 = 0.588581570495;
  const double q2 = 0.531103462366;
  const double q3 = 0.103537752850;
  const double q4 = 0.38560700634e-2;

  double num = (((x*p4+p3)*x+p2)*x+p1)*x+p0;
  double den = (((x*q4+q3)*x+q2)*x+q1)*x+q0;

  // this is our initial guess, which is accurate to about 7 decimal places...
  double x0 = sign*(x+num/den); 

  // ...but the following sharpens our estimate up quite a bit.
  double Z = sqrt(2*M_PI)*(p - normal_cdf(x0))/exp(-x0*x0/2);
  return x0 + (((2*x0*x0 + 1)*Z/3 + x0)*Z/2 + 1)*Z;
}

double
poisson_cdf(double k, double x)
{
  return x <= 0 || k < 1 ? NAN : 1 - incomplete_gamma(k,x);
}

// This function is adapted from one in R (see license header above)
// Are the args screwed up here?  Need to check Hogg & Craig
double
inv_poisson_cdf(double p, double lambda)
{
  if (isnan(p) || isnan(lambda) || p < 0 || p > 1 || lambda <= 0)
    return NAN;

  if (p == 0) return 0;
  if (p == 1) return HUGE_VAL;

  double mu = lambda;
  double sigma = sqrt(lambda);
  double gamma = sigma;
  double z = inv_normal_cdf(p);
  double y = floor(mu + sigma * (z + gamma * (z * z - 1) / 6) + 0.5);
  z = poisson_cdf(y, lambda);
  if (z >= p) {
    /* search to the left */
    for (;;) {
      if ((z = poisson_cdf(y - 1, lambda)) < p)
	return y;
      y = y - 1;
    } 
  } else {
    /* search to the right */
    for (;;) {
      if ((z = poisson_cdf(y + 1, lambda)) >= p)
	return y + 1;
      y = y + 1;
    }
  }
  return -1; // to silence compiler warnings
}

double
chi_square_cdf(double chi2, double df)
{
  return chi2 <= 0 ? 0 : incomplete_gamma(df/2,chi2/2);
}

double
inv_chi_square_cdf(double p, double df)
{
  return inv_gamma_cdf(p,df/2,2);
}

double
students_cdf(double t, double df)
{
  double b = 1 - incomplete_beta(df/2, 0.5, df/(df+t*t));
  return 0.5 + (t >= 0 ? b/2 : -b/2);
}

// This function is adapted from one in R (see license header above)
double
inv_students_cdf(double p, double df)
{
  const double EPSILON = 1e-13;
  double a, b, c, d, prob, P, q, x, y;
  int neg;

  if (isnan(p) || isnan(df) || df < 1 || p > 1 || p < 0)
    return NAN;

  if (p == 0) return -HUGE_VAL;
  if (p == 1) return HUGE_VAL;

  if (df > 1e20)
    return inv_normal_cdf(p);

  if(p > 0.5) {
    neg = 0; P = 2 * (1 - p);
  } else {
    neg = 1; P = 2 * p;
  }

  if (fabs(df - 2) < EPSILON) {
    /* df ~= 2 */
    q = sqrt(2 / (P * (2 - P)) - 2);
  } else if (df < 1 + EPSILON) {
    /* df ~= 1 */
    prob = P * M_PI_2;
    q = cos(prob) / sin(prob);
  } else {
    /*-- usual case;  including, e.g.,  df = 1.1 */
    a = 1 / (df - 0.5);
    b = 48 / (a * a);
    c = ((20700 * a / b - 98) * a - 16) * a + 96.36;
    d = ((94.5 / (b + c) - 3) / b + 1) * sqrt(a * M_PI_2) * df;
    y = pow(d * P, 2 / df);
    
    if (y > 0.05 + a) {
      /* Asymptotic inverse expansion about normal */
      x = inv_normal_cdf(0.5 * P);
      y = x * x;
      if (df < 5)
	c = c + 0.3 * (df - 4.5) * (x + 0.6);
      c = (((0.05 * d * x - 5) * x - 7) * x - 2) * x + b + c;
      y = (((((0.4 * y + 6.3) * y + 36) * y + 94.5) / c - y - 3) / b + 1) * x;
      y = a * y * y;
      if (y > 0.002)
	y = exp(y) - 1;
      else {
	/* Taylor of  e^y -1 : */
	y = 0.5 * y * y + y;
      }
    } else {
      y = ((1 / (((df + 6) / (df * y) - 0.089 * d - 0.822)
		 * (df + 2) * 3) + 0.5 / (df + 4))
	   * y - 1) * (df + 1) / (df + 2) + 1 / y;
    }
    q = sqrt(df * y);
  }
  if (neg) q = -q;
  return q;
}

double
F_cdf(double F, double df1, double df2)
{
  return incomplete_beta(df2/2, df1/2, df2/(df2+df1*F));
}




// $Id: statfn.cc,v 1.7 1998/09/14 04:48:38 trow Exp $
